#!/usr/bin/env python3
"""kmcos.gui.forms - GUI forms used by kmcos.gui
The classes defined here heavily draw on the interface provided by
python-kiwi.
In the language of underlying MVC (Model-View-Controller) pattern these
classes form the controller. The view is defined through a *.glade XML file
and the models are instances of kmcos.types.*
"""
from __future__ import print_function
#    Copyright 2009-2013 Max J. Hoffmann (mjhoffmann@gmail.com)
#    This file is part of kmcos.
#
#    kmcos is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    kmcos is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with kmcos.  If not, see <http://www.gnu.org/licenses/>.
# Standard library imports
import re
import copy
#gtk import
try:
    import pygtk
    pygtk.require('2.0')
    import gtk
    import goocanvas
except ModuleNotFoundError:
    import gi
    gi.require_version('Gtk', '3.0')
    gi.require_version('GooCanvas', '2.0')
    from gi.repository import Gtk as gtk
    from gi.repository import GooCanvas as goocanvas

#kiwi imports
from kiwi.ui.delegates import ProxySlaveDelegate, GladeDelegate, \
                              SlaveDelegate, ProxyDelegate

from kiwi.ui.views import SlaveView
from kiwi.datatypes import ValidationError
from kiwi.ui.objectlist import Column

# own modules
from kmcos.config import GLADEFILE
from kmcos.utils import CorrectlyNamed, \
                       get_ase_constructor, \
                       col_str2tuple, \
                       jmolcolor_in_hex

from kmcos.types import ProcessFormSite, Process, OutputItem, Coord, \
                       ConditionAction, Site

from kmcos import evaluate_rate_expression
from kmcos.types import parse_chemical_expression

# ASE import
import numpy as np
from ase.atoms import Atoms
from ase.data import covalent_radii


class MetaForm(ProxySlaveDelegate, CorrectlyNamed):
    """
    Allows to enter meta information about the project.
    Please enter author and email so people can credit you for the model.

    Increasing the debug level makes the kmcos backed create a lot of
    output but is typically not needed.
    """
    gladefile = GLADEFILE
    toplevel_name = 'meta_form'
    widgets = ['author', 'email', 'model_name', 'model_dimension', 'debug', ]

    def __init__(self, model, project_tree):
        ProxySlaveDelegate.__init__(self, model)
        #self.model_dimension.set_sensitive(False)
        self.project_tree = project_tree
        self.author.set_tooltip_text(
            'Give a name so people know who to credit for the model.')
        self.email.set_tooltip_text(
            'Enter an email address so people can get in touch with you.')
        self.model_name.set_tooltip_text(
            'Give a clear unique name, to identify the model.')
        self.model_dimension.set_tooltip_text(
            'The source code export function can generate ' +
            '1d, 2d, and 3d programs. However this GUI currently only ' +
            'supports 2d. 3d is still possible ' +
            'by manipulating the project XML file by hand. The algorithm ' +
            'though is fast but very memory consuming ' +
            'so a 3d simulation might require considerably more RAM.')
        self.debug.set_tooltip_text(
            'Increasing the debug level might give hints if one suspects ' +
            'errors in kmcos itself. It does not help to debug your model. ' +
            'So usually one wants to keep it a 0.')
        self.author.grab_focus()

    def on_model_name__validate(self, widget, model_name):
        return self.on_name__validate(widget, model_name)

    def on_model_name__content_changed(self, _text):
        self.project_tree.update(self.model)

    def on_model_dimension__content_changed(self, _widget):
        self.project_tree.update(self.model)


class SpeciesListForm(ProxySlaveDelegate):
    """Allows to set the default species, `i. e.` the
    system will be globally initialized with this species if
    nothing else is set on a per site basis.
    """
    gladefile = GLADEFILE
    toplevel_name = 'species_list_form'
    widgets = ['default_species']

    def __init__(self, model, project_tree):
        # this _ugly_ implementation is due to an apparent catch 22 bug in
        # ProxyComboBox: if the value is set already __init__ expect
        # the value in the list but you cannot fill the list before
        # calling __init__
        default_species = model.default_species
        model.default_species = None
        ProxySlaveDelegate.__init__(self, model)
        self.default_species.prefill([x.name
                                      for x in project_tree.get_speciess()],
                                      sort=True)
        self.default_species.select(default_species)
        self.default_species.set_tooltip_text(
            'The lattice will be initialized with this species by default\n'
            + 'but also every unspecified condition or action wil be'
            + 'completed with this choice.\n'
            + 'So better only change this once at the begining if at all!')


class SpeciesForm(ProxySlaveDelegate, CorrectlyNamed):
    """Allows to define a new species. Required attribute is name. The
    chosen color will only shop up in the process editor. So choose something
    you will remember and recognize.
    The representation string is meant to be a ASE ase.atoms.Atoms constructor
    that will show up in the ASE visualization.
    """
    gladefile = GLADEFILE
    toplevel_name = 'species_form'
    widgets = ['name', 'color', 'representation']

    def __init__(self, model, project_tree):
        self.project_tree = project_tree
        ProxySlaveDelegate.__init__(self, model)
        self.name.grab_focus()
        self.name.set_tooltip_text(
        'The name here is arbitrary but you will have to type it many times.'
        + 'So you might want to use e.g. CO instead carbon_monoxide')
        self.color.set_tooltip_text(
            'Choose a color a represent this species in the process editor')
        self.representation.set_tooltip_text(
        'Set an ASE Atoms(\n\'...\') like string to representation in the '
        + 'auto-generated movie. Please only use \'\' for quotation')

    def on_name__content_changed(self, _text):
        self.project_tree.update(self.model)


class ParameterForm(ProxySlaveDelegate, CorrectlyNamed):
    """Allows to set parameter. These
    parameters can be used in e.g. Transition State Theory
    formula to calculate rate constants.

    If 'adjustable' is activated then they maybe be changed via
    the `kmcos view` front end while watching the model run.
    """
    gladefile = GLADEFILE
    toplevel_name = 'parameter_form'
    widgets = ['parameter_name',
               'value',
               'parameter_adjustable',
               'parameter_min',
               'parameter_max']

    def __init__(self, model, project_tree):
        self.project_tree = project_tree
        ProxySlaveDelegate.__init__(self, model)
        value = self.parameter_adjustable.get_active()
        self.parameter_max.set_sensitive(value)
        self.parameter_min.set_sensitive(value)
        self.name.grab_focus()
        self.parameter_adjustable.set_tooltip_text(
        'Settings this adjustable will create a bar in the auto-generated ' +
        'movie. Dragging this bar will adapt the barrier and recalculate ' +
        'all rate constants. This only makes sense for physical ' +
        'parameters such a partial pressure but not for e.g. lattice size')
        self.parameter_name.set_tooltip_text(
        'Choose a sensible name that you remember later when typing rate ' +
        'constant formulae. This should not contain spaces')
        self.value.set_tooltip_text(
            'This defines the initial value for the parameter.')

    def on_parameter_adjustable__content_changed(self, _form):
        value = self.parameter_adjustable.get_active()
        self.parameter_max.set_sensitive(value)
        self.parameter_min.set_sensitive(value)

    def on_value__content_changed(self, _text):
        self.project_tree.update(self.model)

    def on_parameter_name__content_changed(self, _text):
        self.project_tree.update(self.model)
        self.project_tree.project_data.sort_by_attribute('name')


class LatticeForm(ProxySlaveDelegate):
    """Allows to set global lattice parameter such as the lattice vector,
    a ASE representation string, and the default layer. The program will
    be initialized using the default layer.
    """
    gladefile = GLADEFILE
    toplevel_name = 'lattice_form'
    widgets = ['default_layer',
               'lattice_representation']

    def __init__(self, model, dimension, project_tree):
        default_layer = model.default_layer
        model.default_layer = None
        ProxySlaveDelegate.__init__(self, model)
        self.default_layer.prefill([x.name
                                    for x in project_tree.get_layers()],
                                   sort=True)
        self.default_layer.select(default_layer)
        self.default_layer.set_tooltip_text(
        'By default the system will be initialized with this layer.'
        + 'This only matters if using using more than one layer'
        + '(multi-lattice kMC).')

    def on_add_structure__clicked(self, _):
        try:
            import ase.io
        except:
            print('Need ASE to do this.')
            return

        filechooser = gtk.FileChooserDialog(
            title='Open structure file',
            action=gtk.FILE_CHOOSER_ACTION_OPEN,
            buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                     gtk.STOCK_OK, gtk.RESPONSE_OK))
        resp = filechooser.run()
        filename = filechooser.get_filename()
        filechooser.destroy()
        if resp == gtk.RESPONSE_OK and filename:
            try:
                structure = ase.io.read(filename)
                if structure is list:
                    structure = structure[-1]
            except:
                print('Could not open this file. Please choose')
                print('a format that ASE can understand')
                return
            cur_text = self.lattice_representation.get_buffer().get_text(
                self.lattice_representation.get_buffer().get_start_iter(),
                self.lattice_representation.get_buffer().get_end_iter())
            if not cur_text:
                structures = []
            else:
                structures = eval(cur_text)
            structures.append(structure)
            self.lattice_representation.get_buffer().set_text(
                '[%s]' % (
                ', '.join(
                        [get_ase_constructor(x) for x in structures])))


class LayerEditor(ProxySlaveDelegate, CorrectlyNamed):
    """Widget to define a lattice through the sites
    in the unit cell (i.e. the `basis` in solid state language).
    """
    gladefile = GLADEFILE
    toplevel_name = 'layer_form'
    widgets = ['layer_name', 'layer_color']

    def __init__(self, model, project_tree):
        self.project_tree = project_tree
        ProxySlaveDelegate.__init__(self, model)
        self.canvas = goocanvas.Canvas()
        self.root = self.canvas.get_root_item()
        self.canvas.set_size_request(400, 400)
        self.canvas.set_flags(gtk.HAS_FOCUS | gtk.CAN_FOCUS)
        self.canvas.connect('button-press-event', self.on_button_press)

        self.layer_nr = self.project_tree.get_layers().index(model)

        self.radius_scale = 20
        self.scale = 20
        self.offset = np.array([150, 300, 0])
        self.lattice_pad.add(self.canvas)
        self.previous_layer_name = self.layer_name.get_text()
        self.redraw()

        self.layer_name.set_tooltip_text(
            'A name is only relevant if you are using more than one\n'
            + 'layer in your model.')

    def _get_atoms(self):
        if self.project_tree.lattice.representation:
            representations = eval(self.project_tree.lattice.representation)
            if len(representations) > self.layer_nr:
                atoms = representations[self.layer_nr]
            else:
                atoms = representations[0]
        else:
            atoms = Atoms()
        return atoms

    def redraw(self):
        """Draw the current lattice with unit cell
           and sites defined on it.
        """

        # draw atoms in background
        atoms = self._get_atoms()

        self.lower_left = (self.offset[0],
                           self.offset[1]
                           + self.scale * atoms.cell[1, 1])
        self.upper_right = (self.offset[0]
                           + self.scale * atoms.cell[0, 0],
                           self.offset[1])
        big_atoms = atoms * (3, 3, 1)
        for atom in sorted(big_atoms, key=lambda x: x.position[2]):
            i = atom.number
            radius = self.radius_scale * covalent_radii[i]
            color = jmolcolor_in_hex(i)
            X = atom.position[0]
            Y = - atom.position[1]
            goocanvas.Ellipse(parent=self.root,
                              center_x=(self.offset[0] + self.scale * X),
                              center_y=(self.offset[1] + self.scale * Y),
                              radius_x=radius,
                              radius_y=radius,
                              stroke_color='black',
                              fill_color_rgba=color,
                              line_width=1.0)

        # draw unit cell
        A = tuple(self.offset[:2])
        B = (self.offset[0] + self.scale * (atoms.cell[0, 0]),
             self.offset[1] + self.scale * (atoms.cell[0, 1]))

        C = (self.offset[0] + self.scale * (atoms.cell[0, 0]
                                       + atoms.cell[1, 0]),
             self.offset[1] - self.scale * (atoms.cell[0, 1]
                                       + atoms.cell[1, 1]))

        D = (self.offset[0] + self.scale * (atoms.cell[1, 0]),
             self.offset[1] - self.scale * (atoms.cell[1, 1]))
        goocanvas.Polyline(parent=self.root,
                           close_path=True,
                           points=goocanvas.Points([A, B, C, D]),
                           stroke_color='black',)

        # draw sites
        for x in range(3):
            for y in range(3):
                for site in self.model.sites:

                    # convert to screen coordinates
                    pos = np.dot(site.pos + np.array([x, y, 0]), atoms.cell)
                    pos *= np.array([1, -1, 1])
                    pos *= self.scale
                    pos += self.offset
                    X = pos[0]
                    Y = pos[1]

                    o = goocanvas.Ellipse(parent=self.root,
                                          center_x=X,
                                          center_y=Y,
                                          radius_x=.3 * self.radius_scale,
                                          radius_y=.3 * self.radius_scale,
                                          stroke_color='black',
                                          fill_color='white',
                                          line_width=1.0,)

                    o.site = site
                    o.connect('query-tooltip', self.query_tooltip)
                self.canvas.hide()
                self.canvas.show()

    def query_tooltip(self, _canvas, widget, tooltip):
        tooltip.set_text(widget.site.name)
        return True

    def on_button_press(self, _item, event):
        atoms = self._get_atoms()
        pos = (np.array([event.x, event.y, 0]) - self.offset)

        # convert from screen coordinates
        pos *= [1, -1, 1]
        pos /= self.scale
        pos = np.linalg.solve(atoms.cell.T, pos)

        for site in self.model.sites:
            d = np.sqrt((pos[0] - site.pos[0]) ** 2 +
                        (pos[1] - site.pos[1]) ** 2)
            if d < 0.10:
                SiteForm(site, self, self.project_tree, self.model)
                break
        else:
            new_site = Site()
            new_site.name = ''
            new_site.pos = pos

            # Put z position slightly above
            # top atom as a first guess.
            # Assumes a binding distance of 1.3 Angstrom
            atoms = self._get_atoms()
            z_pos = atoms.get_positions()[:, 2]
            z_pos = z_pos if len(z_pos) else [0]
            Z = max(z_pos) + 1.3
            z = Z / atoms.cell[2, 2]
            new_site.pos[2] = z

            new_site.layer = self.model.name
            self.model.sites.append(new_site)
            SiteForm(new_site, self, self.project_tree, self.model)

    def on_layer_name__validate(self, widget, layer_name):
        # TODO: validate lattice name to be unique
        return self.on_name__validate(widget, layer_name)

    def on_layer_name__content_changed(self, widget):
        # Sync layer names in process coords
        new_layer_name = widget.get_text()
        for process in self.project_tree.get_processes():
            for elem in process.condition_list:
                if elem.coord.layer == self.previous_layer_name:
                    elem.coord.layer = new_layer_name
            for elem in process.action_list:
                if elem.coord.layer == self.previous_layer_name:
                    elem.coord.layer = new_layer_name
        self.previous_layer_name = new_layer_name

        self.project_tree.update(self.model)


class SiteForm(ProxyDelegate, CorrectlyNamed):
    """Allows to create or modify a site when setting up a unit cell.
    """
    gladefile = GLADEFILE
    toplevel_name = 'site_form'
    widgets = ['site_name',
               'default_species',
               'site_tags']

    def __init__(self, site, parent, project_tree, layer):
        self.saved_state = copy.deepcopy(site)
        self.project_tree = project_tree
        default_species = site.default_species
        site.default_species = None
        ProxyDelegate.__init__(self, site)

        # fill species dialog with correct available choices
        self.site_default_species.prefill([x.name
                                            for x in
                                            project_tree.get_speciess()],
                                            sort=True)
        if default_species == 'default_species':
            self.site_default_species.select(
                        self.project_tree.species_list.default_species)
        else:
            self.site_default_species.select(default_species)
        self.model.default_species = self.site_default_species.get_selected()

        self.site_ok.grab_default()
        self.site = site

        # set site coordinates
        self.sitevect_x.set_text(str(site.pos[0]))
        self.sitevect_y.set_text(str(site.pos[1]))
        self.sitevect_z.set_text(str(site.pos[2]))

        self.parent = parent
        self.project_tree = project_tree
        self.layer = layer
        self.show_all()
        self.site_name.set_tooltip_text(
            'The site name has to be uniquely identify a site (at least '
            'within each layer for multi-lattice mode). You may have to '
            'type this name a lot, so keep '
            'it short but unambiguous. '
            'To delete a site, erase name.')

    def on_sitevect_x__activate(self, _):
        self.on_site_ok__clicked(_)

    def on_sitevect_y__activate(self, _):
        self.on_site_ok__clicked(_)

    def on_sitevect_z__activate(self, _):
        self.on_site_ok__clicked(_)

    def on_site_name__activate(self, _):
        self.on_site_ok__clicked(_)

    def on_site_name__validate(self, _widget, site_name):
        """check if other site already has the name"""
        if [x for x in self.layer.sites if x.name == site_name]:
            self.site_ok.set_sensitive(False)
            return ValidationError('Site name needs to be unique')
        else:
            self.site_ok.set_sensitive(True)

    def on_site_cancel__clicked(self, _):
        """If we click cancel revert to previous state
        or don't add site, if new."""
        if self.saved_state.name:
            # if site existed, reset to previous state
            self.layer.sites.remove(self.site)
            self.layer.sites.append(self.saved_state)
        else:
            # if site did not exist previously, remove completely
            self.layer.sites.remove(self.site)
        self.hide()
        self.parent.redraw()

    def on_site_ok__clicked(self, _button):
        self.model.default_species = self.site_default_species.get_selected()
        if not len(self.site_name.get_text()):
            self.layer.sites.remove(self.model)
        self.hide()
        self.parent.redraw()


class ProcessForm(ProxySlaveDelegate, CorrectlyNamed):
    """Allows to create and manipulate a process by dragging
    species onto respective sites. The the lower species
    represents a condition the upper one an action.

    Rate constants can be entered directly using all defined parameters.
    The tooltip shows the current value if all is entered correctly.
    """
    gladefile = GLADEFILE
    toplevel_name = 'process_form'
    widgets = ['process_name',
               'rate_constant',
               'process_enabled',
               'chemical_expression']
    z = 5  # z as in zoom
    l = 500  # l as in length
    r_cond = 15.
    r_act = 10.
    r_reservoir = 5.
    r_site = 5.  # where the center unit cell is in the drawing
    X = 2
    Y = 2

    def __init__(self, process, project_tree):
        self.process = process
        self.project_tree = project_tree
        ProxySlaveDelegate.__init__(self, process)
        expression = self.generate_expression()
        self.chemical_expression.update(expression, )

        self.radius_scale = 20
        self.scale = 20
        self.offset = np.array([150, 300, 0])
        self.draw_from_data()

        self.process_name.set_tooltip_text(
        'This name has to uniquely identify the process e.g. co_diff_right')
        self.chemical_expression.set_tooltip_text(
        'This is a fast way to define a process e.g. CO@cus->CO@bridge ' +
        'to declare a CO diffusion from site br to site cus or ' +
        'CO@cus->CO@cus.(0,1) for a CO diffusion in the up direction. Hit ' +
        'ENTER to update the graphical representation.')

        self.rate_constant.curr_value = 0.0
        expr = self.rate_constant.get_text()
        if not expr:
            # if nothing entered show explanation
            self.rate_constant.set_tooltip_text((
                'Python has to be able to evaluate this expression to a ' +
                'plain real number. One can use standard mathematical ' +
                'functions, parameters that are defined under "Parameters"' +
                'or constants and conversion factor such as c, h, e, ' +
                'kboltzmann, pi, bar, angstrom'))
        else:
            try:
                self.rate_constant.set_tooltip_text(
                    'Current value: %.5e s^{-1}' %
                    evaluate_rate_expression(expr,
                        self.project_tree.get_parameters()))
            except Exception as e:
                self.rate_constant.set_tooltip_text(str(e))
        rate_constant_terms = ['bar',
                               'beta',
                               'eV',
                               'exp',
                               'h',
                               'kboltzmann',
                               'umass']
        for param in self.project_tree.get_parameters():
            rate_constant_terms.append(param.name)
        self.rate_constant.prefill(rate_constant_terms)

        chem_exp_terms = ['->', ]
        for species in self.project_tree.get_speciess():
            chem_exp_terms.append(species.name)
        self.chemical_expression.prefill(chem_exp_terms)

    def generate_expression(self):
        expr = ''
        if not self.process.condition_list + self.process.action_list:
            return expr
        for i, condition in enumerate(self.process.condition_list):
            if i > 0:
                expr += ' + '
            expr += '%s@%s' % (condition.species, condition.coord.name)
        expr += ' -> '
        for i, action in enumerate(self.process.action_list):
            if i > 0:
                expr += ' + '
            expr += '%s@%s' % (action.species, action.coord.name)
        return expr

    def on_rate_constant__validate(self, _widget, expr):
        try:
            self.rate_constant.set_tooltip_text('Current value: %.2e s^{-1}' %
                evaluate_rate_expression(expr,
                    self.project_tree.get_parameters()))
        except Exception as e:
            return ValidationError(e)

    def on_chemical_expression__activate(self, entry):
        text = entry.get_text()
        if not text:
            self.process.condition_list = []
            self.process.action_list = []
            self.traw_from_data()
            return
        # Delete trailing plusses
        text = re.sub(r'\s*\+\s', '', text)
        # default to empty right-hand side if not existent
        while text and text[-1] in '-.':
            text = text[:-1]
        if not '->' in text:
            text += '->'
        try:
            parse_chemical_expression(eq=text,
                                      process=self.process,
                                      project_tree=self.project_tree)
            self.process.condition_list = []
            self.process.action_list = []
            parse_chemical_expression(eq=text,
                                      process=self.process,
                                      project_tree=self.project_tree)
        except Exception as e:
            # first remove last term and try again
            try:
                print("Error ...")
                text = re.sub(r'+[^+]*$', '', text)
                parse_chemical_expression(eq=text,
                                          process=self.process,
                                          project_tree=self.project_tree)
                self.process.condition_list = []
                self.process.action_list = []
                parse_chemical_expression(eq=text,
                                          process=self.process,
                                          project_tree=self.project_tree)

            except Exception as e:
                print("Fatal Error ... %s" % e)
                self.process.condition_list = []
                self.process.action_list = []
        finally:
            self.draw_from_data()

    def query_tooltip(self, item, x, y, keyboard_mode, tooltip, *args, **kwargs):
        tooltip.set_text(item.tooltip_text)
        return True

    def on_lattice(self, x, y):
        """Returns True if (x, y) is in lattice box
        """
        return 10 < x < 510 and 80 < y < 580

    def button_press(self, _, item, dummy):
        coords = item.get_coords()
        if item.state == 'reservoir':
            o = CanvasOval(self.motion_layer,
                           *coords, filled=True, bg=item.bg)
            o.connect('button-press-event', self.button_press)
            o.connect('motion-notify-event', self.drag_motion)
            o.connect('button-release-event', self.button_release)
            o.state = 'from_reservoir'
            o.species = item.species
            self.item = o
            self.item.clicked = True
            self.item.father = item
            self.prev_pos = self.item.get_center()
            self.canvas.redraw()

    def drag_motion(self, _widget, _item, event):
        if self.item.clicked:
            d = event.x - self.prev_pos[0], event.y - self.prev_pos[1]
            self.item.move(*d)
            self.prev_pos = event.x, event.y

    #@verbose
    def button_release(self, _, dummy, event):
        self.item.clicked = False
        if self.item.state == 'from_reservoir':
            if not self.on_lattice(event.x, event.y):
                self.item.delete()
            else:
                close_sites = self.site_layer.find_closest(
                                               event.x,
                                               event.y,
                                               halo=(.2 * self.l) / self.z)
                if close_sites:
                    closest_site = min(close_sites,
                                       key=lambda i:
                                        (i.get_center()[0] - event.x) ** 2
                                        + (i.get_center()[1] - event.y) ** 2)
                    coord = closest_site.get_center()
                    self.item.set_center(*coord)
                    if not self.process.condition_list \
                           + self.process.action_list:
                    # if no condition or action is defined yet,
                    # we need to set the center of the editor
                        self.X = closest_site.i
                        self.Y = closest_site.j
                    species = self.item.species
                    offset = closest_site.i - self.X, closest_site.j - self.Y
                    name = closest_site.name
                    layer = closest_site.layer
                    kmc_coord = Coord(offset=offset,
                                  name=name,
                                  layer=layer)
                    condition_action = ConditionAction(species=species,
                                                       coord=kmc_coord)

                    if [x for x in self.condition_layer
                        if x.get_center() == coord]:
                        self.item.new_parent(self.action_layer)
                        self.item.set_radius(self.r_act)
                        self.process.action_list.append(condition_action)
                    else:
                        self.item.new_parent(self.condition_layer)
                        self.item.set_radius(self.r_cond)
                        self.process.condition_list.append(condition_action)
                else:
                    self.item.delete()

        self.chemical_expression.update(self.generate_expression(), )
        self.canvas.redraw()

    def draw_from_data_old(self):
        """Places circles on the current lattice according
        to the conditions and actions defined
        """

        def get_species_color(species):
            return [x for x in self.project_tree.get_speciess()
                    if x.name == species][0].color

        white = col_str2tuple('#ffffff')
        black = col_str2tuple('#000000')
        if hasattr(self, 'canvas'):
            self.process_pad.remove(self.canvas)
        self.canvas = Canvas(bg=white, fg=white)
        self.canvas.set_flags(gtk.HAS_FOCUS | gtk.CAN_FOCUS)
        self.canvas.grab_focus()
        self.canvas.show()
        self.process_pad.add(self.canvas)
        self.lattice_layer = CanvasLayer()
        self.canvas.append(self.lattice_layer)
        self.site_layer = CanvasLayer()
        self.canvas.append(self.site_layer)
        self.condition_layer = CanvasLayer()
        self.canvas.append(self.condition_layer)
        self.action_layer = CanvasLayer()
        self.canvas.append(self.action_layer)
        self.frame_layer = CanvasLayer()
        self.canvas.append(self.frame_layer)
        self.motion_layer = CanvasLayer()
        self.canvas.append(self.motion_layer)

        # draw lattice
        for i in range(self.z):
            CanvasLine(self.lattice_layer,
                       0, i * (self.l / self.z),
                       500, i * (self.l / self.z),
                       line_width=1, fg=(.6, .6, .6))
        for i in range(self.z):
            CanvasLine(self.lattice_layer,
                       i * (self.l / self.z), 0,
                       i * (self.l / self.z), 500,
                       line_width=1, fg=(.6, .6, .6))
        active_layers = [x for x in self.project_tree.get_layers()
                         if x.active]
        site_list = []
        for active_layer in active_layers:
            for site in active_layer.sites:
                form_site = ProcessFormSite(name=site.name,
                                            pos=site.pos,
                                            layer=active_layer.name,
                                            color=active_layer.color)
                site_list.append(form_site)
        for i in range(self.z + 1):
            for j in range(self.z + 1):
                for site in site_list:
                    color = col_str2tuple(site.color)
                    if i == self.X and j == self.Y:
                        l_site = CanvasOval(self.site_layer, 0, 0, 10, 10,
                                                                    fg=color)
                    else:
                        l_site = CanvasOval(self.site_layer, 0, 0, 10, 10,
                                                                    fg=color)

                    l_site.set_center(self.l /
                                      self.z * (i + float(site.pos[0])),
                                      500 - self.l /
                                      self.z * (j + float(site.pos[1])))
                    l_site.connect('query-tooltip', self.query_tooltip)
                    # 500 - ... for having scientific coordinates
                    # and not screen coordinates
                    l_site.set_radius(5)
                    l_site.i = i
                    l_site.j = j
                    if len(active_layers) > 1:
                        l_site.tooltip_text = '%s.(%s,%s).%s' % (site.name,
                                                                 i - self.X,
                                                                 j - self.Y,
                                                                 site.layer)
                    else:
                        l_site.tooltip_text = '%s.(%s,%s)' % (site.name,
                                                              i - self.X,
                                                              j - self.Y)
                    l_site.name = site.name
                    l_site.offset = (i - self.X, j - self.Y)
                    l_site.layer = site.layer

        # draw frame
        frame_col = (.21, .35, .42)
        CanvasRect(self.frame_layer, 0, 0, 520, 80, fg=frame_col,
                                                    bg=frame_col,
                                                    filled=True)
        CanvasRect(self.frame_layer, 0, 0, 10, 580, fg=frame_col,
                                                    bg=frame_col,
                                                    filled=True)
        CanvasRect(self.frame_layer, 510, 0, 520, 580, fg=frame_col,
                                                       bg=frame_col,
                                                       filled=True)
        CanvasRect(self.frame_layer, 0, 580, 520, 590, fg=frame_col,
                                                       bg=frame_col,
                                                       filled=True)
        CanvasText(self.frame_layer, 10, 10, size=8, text='Reservoir Area')
        CanvasText(self.frame_layer, 10, 570, size=8, text='Lattice Area')

        # draw reservoir circles
        for k, species in enumerate(self.project_tree.get_speciess()):
            color = col_str2tuple(species.color)
            o = CanvasOval(self.frame_layer,
                           30 + k * 50,
                           30, 50 + k * 50,
                           50,
                           filled=True,
                           bg=color)
            o.species = species.name
            o.tooltip_text = species.name  # for tooltip
            o.connect('button-press-event', self.button_press)
            #o.connect('motion-notify-event', self.drag_motion)
            o.connect('button-release-event', self.button_release)
            o.connect('query-tooltip', self.query_tooltip)
            o.state = 'reservoir'

        self.lattice_layer.move_all(10, 80)
        self.site_layer.move_all(10, 80)

        # attributes need for moving objects
        self.item = None
        self.prev_pos = None
        black = col_str2tuple('#003333')
        for elem in self.process.condition_list:
            matching_sites = [x for x in self.site_layer
                              if isinstance(x, CanvasOval)
                              and x.i == self.X + elem.coord.offset[0]
                              and x.j == self.Y + elem.coord.offset[1]
                              and x.name == elem.coord.name
                              and x.layer == elem.coord.layer]
            if matching_sites:
                coords = matching_sites[0].get_coords()
                color = get_species_color(elem.species)
                color = col_str2tuple(color)
                o = CanvasOval(self.condition_layer,
                               bg=color,
                               fg=black,
                               filled=True, outline=True)
                o.coords = coords
                o.connect('button-press-event',
                          self.on_condition_action_clicked)
                o.set_radius(self.r_cond)
                o.type = 'condition'
                o.condition = elem
                o.tooltip_text = '%s@%s' % (elem.species, elem.coord)  # for tooltip
                o.connect('query-tooltip', self.query_tooltip)

        for elem in self.process.action_list:
            matching_sites = [x for x in self.site_layer
                              if isinstance(x, CanvasOval)
                              and x.i == self.X + elem.coord.offset[0]
                              and x.j == self.Y + elem.coord.offset[1]
                              and x.name == elem.coord.name
                              and x.layer == elem.coord.layer]
            if matching_sites:
                coords = matching_sites[0].get_coords()
                if elem.species[0] == '^':
                    color = get_species_color(elem.species[1:])
                    layer = self.action_layer
                    radius = self.r_act
                    line_width = 2.0
                elif elem.species[0] == '$':
                    color = get_species_color(elem.species[1:])
                    layer = self.condition_layer
                    radius = self.r_cond
                    line_width = 2.0
                else:
                    color = get_species_color(elem.species)
                    layer = self.action_layer
                    radius = self.r_act
                    line_width = 1.0
                color = col_str2tuple(color)
                o = CanvasOval(layer,
                               bg=color,
                               fg=black,
                               line_width=line_width,
                               filled=True,
                               outline=True)
                o.coords = coords
                o.connect('button-press-event',
                          self.on_condition_action_clicked)
                o.set_radius(radius)
                o.type = 'action'
                o.action = elem
                o.tooltip_text = '%s@%s' % (elem.species, elem.coord)  # for tooltip
                o.connect('query-tooltip', self.query_tooltip)

    def draw_from_data(self):
        atoms = self._get_atoms()
        def toscrn(coord,
                   screen_size=(500, 500),
                   scale=None,
                   offset=None):

            if scale is None:
                scale = min(screen_size[0]/(atoms.cell[0] + atoms.cell[1])[0],
                            screen_size[1]/(atoms.cell[0] + atoms.cell[1])[1])
                scale /= (zoom + 1)

            if offset is None:
                offset = ((screen_size[0] - zoom*scale*(atoms.cell[0] + atoms.cell[1])[0])/2,
                          (screen_size[1] - zoom*scale*(atoms.cell[0] + atoms.cell[1])[1])/2,)
            return (scale * coord[0] + offset[0],
                    screen_size[1] - (scale * coord[1] + offset[1]))

        # automatically determine zoom from process list
        zoom = 2 * self.process._get_max_d() + 3

        center_x = zoom / 2
        center_y = zoom / 2
        if hasattr(self, 'canvas'):
            self.process_pad.remove(self.canvas)
        canvas = goocanvas.Canvas()
        self.canvas = canvas
        root = canvas.get_root_item()
        canvas.set_flags(gtk.HAS_FOCUS | gtk.CAN_FOCUS)
        canvas.set_property('has-tooltip', True)
        #canvas.grab_focus()
        canvas.show()
        self.process_pad.add(canvas)

        radius = 10

        # draw lattice
        for i in range(zoom + 1):
            for _0, _1, _2, _3 in [[i, i, 0, zoom],
                                   [0, zoom, i, i]]:
                points = goocanvas.Points([
                    toscrn(atoms.cell[0]*_0 + atoms.cell[1]*_2),
                    toscrn(atoms.cell[0]*_1 + atoms.cell[1]*_3),
                ])
                goocanvas.Polyline(parent=root,
                                  points=points,
                                  stroke_color='black',
                                  fill_color='white',
                                  line_width=1.0)
        # emphasize central cell
        points = goocanvas.Points([
            toscrn(atoms.cell[0]*center_x + atoms.cell[1]*center_x),
            toscrn(atoms.cell[0]*center_x + atoms.cell[1]*(center_x + 1)),
            toscrn(atoms.cell[0]*(center_x + 1) + atoms.cell[1]*(center_x + 1)),
            toscrn(atoms.cell[0]*(center_x + 1) + atoms.cell[1]*center_x),
            toscrn(atoms.cell[0]*center_x + atoms.cell[1]*center_x),
        ])
        goocanvas.Polyline(parent=root,
                          points=points,
                          stroke_color='black',
                          fill_color='white',
                          line_width=2.0)
        # draw sites
        for x in range(zoom):
            for y in range(zoom):
                sites = self.project_tree.get_layers()[0].sites
                for site in sites:
                    X, Y = toscrn(x*atoms.cell[0]
                                  + y*atoms.cell[1]
                                  + atoms.cell[0] * site.pos[0]
                                  + atoms.cell[1] * site.pos[1]
                                  #+ np.inner(atoms.cell.T, site.pos)
                                  )
                    tooltip = '%s.(%s, %s, 0).%s' % (site.name,
                                                     x-center_x, y-center_y,
                                                     self.project_tree.get_layers()[0].name
                                                     )

                    o = goocanvas.Ellipse(parent=root,
                                          center_x=X,
                                          center_y=Y,
                                          radius_x=.4 * radius,
                                          radius_y=.4 * radius,
                                          stroke_color='black',
                                          fill_color='white',
                                          line_width=1.0,
                                          tooltip=tooltip,
                                          )

        # draw reservoir circles
        offset = np.array([1, 1, 0])
        offset = atoms.cell[0] * center_x + atoms.cell[1] * center_y
        for k, species in enumerate(self.project_tree.get_speciess()):
            color = col_str2tuple(species.color)
            o = goocanvas.Ellipse(parent=root,
                           center_x=30 + k * 50,
                           center_y=30,
                           radius_x=0.8*radius,
                           radius_y=0.8*radius,
                           stroke_color='black',
                           fill_color_rgba=eval('0x' + species.color[1:] + 'ff' ),
                           tooltip=species.name,
                           )

        for elem in self.process.condition_list:
            pos = [x.pos
                   for layer in self.project_tree.get_layers()
                   for x in layer.sites
                   if x.name == elem.coord.name
                      ][0]
            species_color = [x.color for x in self.project_tree.get_speciess()
                             if x.name == elem.species.split(' or ')[0]][0]
            center = toscrn(pos[0] * atoms.cell[0]
                            + pos[1] * atoms.cell[1]
                            #np.inner(atoms.cell, pos)
                            + elem.coord.offset[0] * atoms.cell[0]
                            + elem.coord.offset[1] * atoms.cell[1]
                            + offset)

            tooltip = 'Condition: %s@%s.%s.%s' % (elem.species,
                                       elem.coord.name,
                                       tuple(elem.coord.offset),
                                       elem.coord.layer,
                                       )  # for tooltip
            o = goocanvas.Ellipse(parent=root,
                                  center_x=center[0],
                                  center_y=center[1],
                                  radius_x=0.8*radius,
                                  radius_y=0.8*radius,
                                  stroke_color='black',
                                  fill_color_rgba=eval('0x' + species_color[1:] + 'ff' ),
                                  tooltip=tooltip,
                                  )


        for elem in self.process.action_list:
            species_color = [x.color for x in self.project_tree.get_speciess()
                             if x.name == elem.species][0]
            pos = [x.pos
                   for layer in self.project_tree.get_layers()
                   for x in layer.sites
                   if x.name == elem.coord.name
                      ][0]

            center = toscrn(pos[0] * atoms.cell[0]
                            + pos[1] * atoms.cell[1]
                            + elem.coord.offset[0] * atoms.cell[0]
                            + elem.coord.offset[1] * atoms.cell[1]
                            + offset)

            tooltip = 'Action: %s@%s.%s.%s' % (elem.species,
                                       elem.coord.name,
                                       tuple(elem.coord.offset),
                                       elem.coord.layer)  # for tooltip

            o = goocanvas.Ellipse(parent=root,
                                  center_x=center[0],
                                  center_y=center[1],
                                  radius_x=0.4*radius,
                                  radius_y=0.4*radius,
                                  stroke_color='black',
                                  fill_color_rgba=eval('0x' + species_color[1:] + 'ff' ),
                                  tooltip=tooltip,
                                  )

        # For otf backend only
        if self.process.bystander_list:
            for elem in self.process.bystander_list:
                species_color = '#d3d3d3'
                pos = [x.pos
                    for layer in self.project_tree.get_layers()
                    for x in layer.sites
                    if x.name == elem.coord.name
                        ][0]

                center = toscrn(pos[0] * atoms.cell[0]
                                + pos[1] * atoms.cell[1]
                                + elem.coord.offset[0] * atoms.cell[0]
                                + elem.coord.offset[1] * atoms.cell[1]
                                + offset)
                tooltip = 'Bystander (%s): %s@%s.%s.%s' % (elem.flag,
                                                              elem.allowed_species,
                                                              elem.coord.name,
                                                              tuple(elem.coord.offset),
                                                              elem.coord.layer)  # for tooltip
                bystander_size_factor = 1.2
                o = goocanvas.Rect(parent=root,
                                    x=center[0]-0.6*radius,
                                    y=center[1]-0.6*radius,
                                    width=bystander_size_factor*radius,
                                    height=bystander_size_factor*radius,
                                    stroke_color='black',
                                    fill_color_rgba=eval('0x' + species_color[1:] + 'ff' ),
                                    tooltip=tooltip,
                                    )



    def _get_atoms(self, layer_nr=0):
        if self.project_tree.lattice.representation:
            representations = eval(self.project_tree.lattice.representation)
            if len(representations) > layer_nr:
                atoms = representations[layer_nr]
            else:
                atoms = representations[0]
        else:
            atoms = Atoms()
        return atoms

    def on_condition_action_clicked(self, _canvas, widget, event):
        if event.button == 2:
            if widget.type == 'action':
                self.process.action_list.remove(widget.action)
            elif widget.type == 'condition':
                self.process.condition_list.remove(widget.condition)
            widget.delete()

    def on_process_name__content_changed(self, _text):
        self.project_tree.project_data.sort_by_attribute('name')
        self.project_tree.update(self.process)

    def on_rate_constant__content_changed(self, _text):
        self.project_tree.update(self.process)


class BatchProcessForm(SlaveDelegate):
    """Allows to enter many processes at once. The format is one
    process per line in the form::

        [process name] ; [chemical expression] ; [rate constant]

    One can omit the fields but not the semicolon.
    """
    gladefile = GLADEFILE
    toplevel_name = 'batch_process_form'

    def __init__(self, project_tree):
        self.project_tree = project_tree
        SlaveDelegate.__init__(self)

    def on_btn_evaluate__clicked(self, _):
        batch_buffer = self.batch_processes.get_buffer()
        bounds = batch_buffer.get_bounds()
        text = batch_buffer.get_text(*bounds)
        text = text.split('\n')
        for i, line in enumerate(text):
            # Ignore empty lines
            if not line.count(';'):
                continue
            if not line.count(';'):
                raise UserWarning(
                    ("Line %s: the number of fields you entered is %s, " \
                     "but I expected 3") % (i, line.count(';') + 1))
            line = line.split(';')
            name = line[0]
            if len(line) == 1:
                rate_constant = ''
            elif len(line) == 2:
                rate_constant = ''
            elif len(line) == 3:
                rate_constant = line[2]
            else:
                raise UserWarning(
                    "There are too many ';' in your expression %s" % line)
            process = Process(name=name, rate_constant=rate_constant)
            try:
                parse_chemical_expression(eq=line[1],
                                          process=process,
                                          project_tree=self.project_tree)
                self.draw_from_data()
            except:
                raise Exception(
                    ("Found an error in your chemical expression(line %s):\n"\
                    "%s") % (i + 1, line[1]))
            else:
                # replace any existing process with identical names
                for dublette_proc in [x for x in
                                      self.project_tree.process_list
                                      if x.name == name]:
                    self.project_tree.process_list.remove(dublette_proc)
                self.project_tree.append(self.project_tree.process_list_iter,
                                                                     process)
        batch_buffer.delete(*bounds)


class OutputForm(GladeDelegate):
    """Not implemented yet
    """
    gladefile = GLADEFILE
    toplevel_name = 'output_form'
    widgets = ['output_list']

    def __init__(self, output_list, project_tree):
        GladeDelegate.__init__(self)
        self.project_tree = project_tree
        self.output_list_data = output_list
        self.output_list.set_columns([Column('name',
                                          data_type=str,
                                          editable=True, sorted=True),
                                      Column('output',
                                          data_type=bool,
                                          editable=True)])

        for item in self.output_list_data:
            self.output_list.append(item)

        self.output_list.show()
        self.output_list.grab_focus()

    def on_add_output__clicked(self, _):
        output_form = gtk.MessageDialog(parent=None,
                                      flags=gtk.DIALOG_MODAL,
                                      type=gtk.MESSAGE_QUESTION,
                                      buttons=gtk.BUTTONS_OK_CANCEL,
                                      message_format='Please enter a new ' \
                                         + 'output: examples are a species ' \
                                         + 'or species@site')
        output_form.set_flags(gtk.CAN_DEFAULT | gtk.CAN_FOCUS)
        output_form.set_default_response(gtk.RESPONSE_OK)
        output_form.set_default(
                        output_form.get_widget_for_response(gtk.RESPONSE_OK))
        form_entry = gtk.Entry()

        def activate_default(_):
            output_form.activate_default()
        form_entry.connect('activate', activate_default)
        output_form.vbox.pack_start(form_entry)
        output_form.vbox.show_all()
        res = output_form.run()
        output_str = form_entry.get_text()
        output_form.destroy()
        if res == gtk.RESPONSE_OK:
            output_item = OutputItem(name=output_str, output=True)
            self.output_list.append(output_item)
            self.output_list_data.append(output_item)


class InlineMessage(SlaveView):
    """Return a nice little field with a text message on it
    """
    gladefile = GLADEFILE
    toplevel_name = 'inline_message'
    widgets = ['message_label']

    def __init__(self, message=''):
        SlaveView.__init__(self)
        self.message_label.set_text(message)



#!/usr/bin/env python3
"""
    Several commonly used constants and conversion factors are offered
    by this module, such a pi, the speed of light, the Planck constant,
    the charge of an electron, k_Boltzmann, the atomic masses of carbon
    and oxygen, the mass of a nucleon, one angstrom in meter, and one
    bar in Pascal.
    Source: CODATA2010
"""
#    Copyright 2009-2013 Max J. Hoffmann (mjhoffmann@gmail.com)
#    This file is part of kmcos.
#
#    kmcos is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    kmcos is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with kmcos.  If not, see <http://www.gnu.org/licenses/>.
keys = ['pi', 'c', 'h', 'hbar', 'eV', 'kboltzmann',
        'umass', 'angstrom', 'bar']

pi = 3.14159265358979323846  # approximately ...
c = 2.99792458e8  # m/s
h = 6.62606957e-34  # J s
hbar = 1.054571726e-34  # J s
eV = 1.602176565e-19  # C
kboltzmann = 1.3806488e-23  # J K
umass = 1.660538921e-27  # kg atomic mass
angstrom = 1.E-10  # m
bar = 1.E5  # kg / m s^2



from kmc_settings import simulation_size

# Be sure to use floats and not integers in general, unless integers make sense
# (e.g., process indices).

################################################################################
#                                                                              #
#                  PART I: USER-ADJUSTABLE/DEFINED PARAMETERS                  #
#                                                                              #
################################################################################

# Number of unit cells in the simulation box
Nsites = simulation_size**2

# Should we print detailed information about the throttling?
print_throttling_info = True

# This is the currently chosen ranking scheme. We have this variable to permit
# flexibly changing the ranking scheme on-the-fly if desired. Currently the only
# type supported is 'paired', but other types may be implemented in the future.
current_ranking_scheme = 'paired'

# A regular expression for extracting only the relevant processes from a list of
# processes created by the snapshots module. Note that the regex r'.*' will
# match everything, thereby allowing one to use the raw TOF lists directly.
#regex = r'.*' # uncomment to turn off regular expression matching
regex = r'^[FR]\d+p\d+$' # comment out if turning off regex matching

# Factor controlling whether reactions are considered to be quasi-equilibrated
# or not. Current default is 0.1.
qe_tolerance = 0.1

# Flag controlling the compression progression employed. The tag 'auto' means to
# use a default progression up to a requested compression range. Otherwise, the
# scheme should be specfied as a list of 2-tuples, with the first (second)
# element of each tuple specifying the fast (slow) compression amount. The
# Python object None can be used to specify no compression is allowed.
# Examples:
#   compression_scheme_flag = 'auto' # default progression
#   compression_scheme_flag = None # No compression, sets compression_schemes = [(0, 0)]
#   compression_scheme_flag = 'manual' # Manual setting of compression schemes
#
# If 'manual' is selected, then set the compression_schemes variable as a list
# of 2-tuples with the fast and slow scaling levels, as in:
#   compression_schemes = [(1, 0)] # Nsites spacing between all fast processes
#   compression_schemes = [(1, 0), (0, 1)] # Nsites spacing between all fast
#     processes followed by Nsites spacing between all slow processes.
# The variable compression_schemes will be overwritten if auto or none is
# selected in the flag.
compression_scheme_flag = 'auto'
compression_schemes = None

# Maximum throttling levels for fast and slow scales. The maximum allowed value
# for this is 3 for each one. If this is set to zero, then no throttling will be
# allowed for that scale.
max_fast_scale = 3
max_slow_scale = 3

# Whether to throttle slow and fast compression ranges separately ('split') or
# together ('full'); 'split' should usually be used.
EF_range_flag = 'split'

# Requested compression ranges (ratio of fastest rate to slowest rate). These
# are adjusted below depending on the value of EF_range_flag. Any ranges not
# associated with the specified EF_range_flag are set to None.
EF_range_fast_requested = 1e2     # For fast processes only ('split')
EF_range_slow_requested = 1e20    # For slow processes only ('split')
EF_range_full_requested = 1e6     # For both fast and slow together ('full')

# Any processes not expected to occur more than Nsites times within this cutoff
# time during the simulations will be excluded from throttling as irrelevant.
max_time = 10. * 3600 #This is the maximum relevant simulation time (MRST in the original SQERTSS paper)

# This is the maximum relevant simulation time that we are concerned with
# simulating. This is a hard limit -- if we reach it, the throttling loop will
# stop.
cutoff_time = 100.0

# The characteristic_EF is the primary benchmark for FFPs. In a system with FFPs, the throttling
# algorithm will step down the slowest FFP until it approaches this value. The value
# is user-adjustable and based on the desired timescale. The floor should be
# approximately the rate of the FRP times Nsites. We leave this in the
# user-adjustable section as this is just a guideline and can be changed as
# desired. If the user wants to change this value, it should be done in the
# runfile.
characteristic_EF = 0.2  # Characteristic transition rate
FFP_floor = characteristic_EF * Nsites
#it is good to use an FFP_roof: the compression will be required to satisfy *both* the roof *and* the EF_range requirementes.
FFP_roof = None 
#By default, the FRP_Buffer between FFP_floor and the FRP is Nsites.
FRP_Buffer = Nsites 

# Total number of KMC steps to execute per snapshot. This should be at least
#   throttling_sps = EF_range_fast_requested * Nsites * n_characteristic_events_target
# for split range and
#   throttling_sps = EF_range_full_requested
# for full range. If the user wants to change this value, it should be done
# here or in the runfile. The variable n_characteristic_events_target is the
# number of times that the characteristic event (e.g., FRP) should occur during
# the snapshot.
n_characteristic_events_target = 10
if EF_range_flag == 'split':
    throttling_sps = int(EF_range_fast_requested * Nsites *
        n_characteristic_events_target)
else:
    throttling_sps = int(EF_range_full_requested)

# Total KMC time to execute per snapshot. A value of None will cause the
# algorithm to use a fixed number of steps (this should be the typical case).
throttling_tps = None

# Default FFP step down divisor. A smaller divisor will step down less, and any
# value less than one will actually increase the FFP rate (*so don't use a value
# less than 1*). This is used in conjunction with the number of KMC steps and
# the number of sites to automatically choose a reasonable down step divisor. A
# value of 10 is very safe (conservative). The guideline of
# throttling_sps/Nsites is more aggressive but still reasonable, at least in the
# case where fixed step snapshots are employed (time step snapshots are still a
# work in progress, and the default should be used).
default_FFP_step_down = 10.0

# Do we want to use guidelines for setting various quantities?
use_guideline_FFP_step_down = True  # FFP step down
use_guideline_FFP_floor = False  # Floor level #this will override what is set for FFP_floor
use_guideline_sps = False    # SPS size #this will override what is set for throttling_sps

# These are the staggering factors controlling the compression of scales. Each
# factor is used at the corresponding scale (1, 2, or 3). These are all
# reasonable values and are user-adjustable, but any adjustments should be made
# with care! Note that we also have Nsites for the initial entry. This entry is
# only used during FFP step down, and it ensures that the sFFP will not get
# closer than Nsites to an FQP if it exists.
staggering_factors = [Nsites, Nsites, 10., 1.1]

# Total number of steps to take between unthrottling of slow reactions
loop_base = 10 #This is "M" in the original SQERTSS paper.

# If the current snapshot counter is an integer multiple of loop_base, should we
# unthrottle all processes (not just the slow ones) by resetting the local
# snapshot counter to zero?
reset_snapshot_counter = True

# Throttle forward/reverse process pairs together?
throttle_process_pairs = True

# Should we enforce steady state throttling where slow processes are never
# throttled?
steady_state_throttling = False

# General numerical tolerance to allow some slack in determining equality of two
# close floating point numbers. If a calculated quantity is larger than the
# expected amount or requested criterion (e.g., in the EF_range compression
# ratios) by less than this amount, it is treated as still meeting the test. We
# have this to permit some (small) discrepancies to pass that arise from finite
# numerical precision. This tolerance is relative to the requested/expected
# value. It should be fairly small.
floating_point_rel_tol = 1e-4

# Should we regularize rate constants that are too large?
regularization = False

################################################################################
#                                                                              #
#                  PART II: AUTOMATICALLY GENERATED VARIABLES                  #
#                           !! DO NOT EDIT THESE !!                            #
#                                                                              #
################################################################################

# This is a list of all elementary processes in the model. It is a subset of the
# TOF header array in snapshots_globals, but is constructed in the runfile to
# eliminate a dependency.
proc_names = []

# This is a list of the observed event frequencies from the current snapshot
# corresponding to the process names list. It is also created in the runfile.
oEF_TOF_list_next = []

# This is a list of the observed event frequencies from the last snapshot.
oEF_TOF_list = []

# This is a list of paired indices for the forward and reverse processes
# corresponding to each reaction number. This dictionary is populated in
# uEF_back_calculate by a call to a helper function.
EF_indices_dict = {}

# The following list-of-lists is the raw, unsorted process rate data. It can be
# used to generate the sublists for the ranked dictionary variables. Each
# sublist has seven elements: reaction number (e.g., 15p0), forward rate,
# reverse rate, ratio of forward to reverse rate, maximum of forward and reverse
# rate, the direction of the maximum rate ('F' or 'R' for forward/reverse), and
# an index indicating the original order before any ranking is performed.
uEF_list = []

# This list is similar to uEF_list, but it has the projected throttled rates.
ptEF_list = []

# These variables contain the current index values for the benchmark and
# slowest rate limiting processes. These indices can be used to extract various
# quantities from the ranked_EF lists.
BP_index = 0
sSP_index = 0

# Dictionaries with the names and types of the current benchmark process
BP_name = None
BP_type = None

# This is a list collecting the processes sorted according to various
# schemes. The format for each scheme is identical. Each scheme has a single
# list with an arbitrary number of related three element sublists. Each sublist
# collects the process number (e.g., F15p0), the event frequency, and a 2-tuple
# with (1) the index to the reaction in uEF_list and (2) the index to the
# forward/reverse process in the process list in uEF_list. The first three
# element sublist is always the process that is deemed most important for the
# grouping. For example, for the paired rank scheme, it is the process with the
# maximum rate.
ranked_uEF_list = []

# These lists contain the ascending FFP and descending SP lists for the current
# ranking scheme. Each of the sublists has the reaction number, associated rate,
# direction of maximum rate, and a 2-tuple with indices to its location in the
# master uEF_list structure and its location in the ranked_uEF_list. The rates
# in these lists are unthrottled.
ascend_FFP_list = []
descend_SP_list = []

# The following variables are all used in the throttling functions.

# These are the aggregate and incremental throttling factors lists. They are
# used to store the total throttling over all snapshots (aggregate factors) and
# new throttling due to this snapshot (incremental factors). The aggregate
# factors are used to back-calculate unthrottled rates that are used to rank
# processes for determining the new incremental throttling factors. These lists
# have an exact correspondence to the ascend_FFP and descend_SP lists.
#
# The structure of these lists is as follows:
#   First index: [0] - list of FFPs (ascending order), [1] - list of SPs
#       (descending order)
#   Second index (for each FFP/SP): Selects sublist for individual processes
#   Third index: [0] - reaction number, [1] - throttling factor,
#       [2] - reaction direction, [3] - process name (F/R + reaction number),
#       [4] - 2-tuple with (1) index to reaction in uEF_list and (2) index to
#       corresponding entry in ranked_uEF_list.
aggregate_throttling_factors = [[], []]
incremental_throttling_factors = [[], []]

# These lists contain the names of the never throttled and previously throttled
# negligible processes.
unthrottled_negligible_process_names = [] # never throttled
throttled_negligible_process_names = [] # previously throttled

# This list contains the names of the fast and slow processes that have been
# throttled. It is used to undo the throttling for only the throttled processes
# during the unthrottling step. This list is also not sorted in any particular
# order.
#
# The structure of this list is as follows:
#   First index: [0] - list of FFPs, [1] - list of SPs
#   Second index: list of process names
throttled_process_names = [[],[]]

# Amount by which to throttle the slowest quasi-equilibrated step in a system
# with only quasi-equilibrated steps (i.e., only FFPs). A value of 10 is very
# safe (conservative); a value of throttling_sps/Nsites is more agressive but
# still reasonable. This approximation is only valid for fixed-step snapshot
# execution and needs to be revisited for time-step snapshot execution. Here we
# choose 10 as the minimum amount allowed, but this is user adjustable. A value
# closer to 1 means less down throttling, and a value less than 1 will actually
# speed up the slowest FFP. (*So don't go below 1!*)
if use_guideline_FFP_step_down:
    FFP_step_down = max(float(throttling_sps)/Nsites, default_FFP_step_down)
else:
    FFP_step_down = default_FFP_step_down

# Variable containing the kind of FFP step down behavior
FFP_step_down_type = None

# List containing the back-calculated unthrottled event frequencies (i.e., the
# rates at which each reaction would have occurred had throttling not been
# applied). This is important for periodically recalibrating the throttling.
# The structure is a simple list with the same order as sg.atoms.tof_data.
uEF_TOF_list = []

# Current aggregate throttling factors for all processes ever throttled. The
# format is a simple dictionary where keys are process names (e.g., F15p0) and
# the values are the current aggregate throttling factors.
aggregate_throttling_factors_dict = {}

# Aggregate throttling factors for the last snapshot
aggregate_throttling_factors_dict_old = {}

# The format (a simple dictionary) is the same as for
# aggregate_throttling_factors_dict, but it contains the incremental throttling
# factors for this snapshot only. It is currently only used for debugging
# purposes.
incremental_throttling_factors_dict = {}

# Variables containing the current kind of compression applied at each level
fast_throttling_scale = 0
slow_throttling_scale = 0

# Adjustments to the requested EF range values
if EF_range_flag == 'full':
    EF_range_fast_requested = None
    EF_range_slow_requested = None
elif EF_range_flag == 'split':
    EF_range_full_requested = None
else:
    print('WARNING: Invalid EF_range_flag detected')
    EF_range_fast_requested = None
    EF_range_slow_requested = None
    EF_range_full_requested = None

# Variables containing the actual compression ratios
EF_range_fast_actual = None
EF_range_slow_actual = None
EF_range_full_actual = None

# Round-off error of the machine; this value is true for CPython on standard x86
# hardware for IEEE double precision floating point values (the standard float
# in Python on all or nearly all platforms is IEEE double precision). This is
# the smallest quantity that when added to 1 results in a sum that is
# distinguishible from 1.
machine_epsilon = 2**-52

# Number of resolvable decimal digits at the cutoff time
resolvable_digits = 2

# Maximum rate constant allowed. Any processes with larger rate constants will
# be capped at this value. This quantity should be chosen with care. If any rate
# constant is too large, the amount of time resolvable by the KMC simulation
# will drop below numerical machine precision and calculated TOF values will
# become undefined. The default value here is sufficient to permit ~3 decimal
# digits of resolution in the KMC time increment at the end of the simulation.
max_rate_constant = 1./(max_time / Nsites * machine_epsilon * 10**resolvable_digits)

# Current snapshot number
current_snapshot = 0

# Original values of pre-exponentials
preexp_dict_original = None # Current snapshot
preexp_dict_original_reg_old = None # Regularized pre-exponentials from the last snapshot
modified_preexp_set = set() # Set of all processes with pre-exponentials that have been modified

# NSP threshold -- below this processes will be NSPs
NSP_EF_threshold = float(Nsites)/max_time

# Flag whether we have exceeded the cutoff time
cutoff_time_exceeded = False

# List of variables to save. *DO NOT EDIT THIS UNLESS YOU KNOW WHAT YOU ARE
# DOING!*
__var_list__ = ['print_throttling_info', 'aggregate_throttling_factors',
    'aggregate_throttling_factors_dict',
    'aggregate_throttling_factors_dict_old', 'ascend_FFP_list', 'BP_index',
    'BP_name', 'BP_type', 'staggering_factors', 'compression_schemes',
    'compression_scheme_flag', 'current_ranking_scheme', 'current_snapshot',
    'cutoff_time', 'default_FFP_step_down', 'descend_SP_list', 'proc_names',
    'EF_indices_dict', 'oEF_TOF_list_next', 'oEF_TOF_list', 'uEF_list',
    'ptEF_list', 'EF_range_fast_requested', 'EF_range_fast_actual', 'EF_range_flag',
    'EF_range_full_requested', 'EF_range_full_actual', 'EF_range_slow_requested',
    'EF_range_slow_actual', 'fast_throttling_scale', 'FFP_floor',
    'FFP_step_down', 'FFP_step_down_type', 'incremental_throttling_factors',
    'incremental_throttling_factors_dict', 'loop_base', 'machine_epsilon',
    'max_rate_constant', 'max_time', 'Nsites', 'preexp_dict_original',
    'preexp_dict_original_reg_old', 'qe_tolerance', 'ranked_uEF_list',
    'resolvable_digits', 'slow_throttling_scale', 'sSP_index',
    'steady_state_throttling', 'reset_snapshot_counter',
    'throttled_negligible_process_names', 'throttled_process_names',
    'throttle_process_pairs', 'throttling_sps', 'uEF_TOF_list',
    'unthrottled_negligible_process_names', 'use_guideline_FFP_step_down',
    'floating_point_rel_tol', 'modified_preexp_set', 'regularization']



#!/usr/bin/env python3
"""Entry point module for the command-line
   interface. The kmcos executable should be
   on the program path, import this modules
   main function and run it.

   To call kmcos command as you would from the shell,
   use ::

       kmcos.cli.main('...')

   Every command can be shortened as long as it is non-ambiguous, e.g. ::


    kmcos ex <xml-file>

   instead of ::

    kmcos export <xml-file>


   etc.

"""

#    Copyright 2009-2013 Max J. Hoffmann (mjhoffmann@gmail.com)
#    This file is part of kmcos.
#
#    kmcos is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    kmcos is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with kmcos.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import with_statement
from __future__ import print_function
import os
import shutil

usage = {}
usage['all'] = """kmcos help all
    Display documentation for all commands.
                """
usage['benchmark'] = """kmcos benchmark
    Run 1 mio. kMC steps on model in current directory
    and report runtime.
                     """

usage['build'] = """kmcos build
    Build kmc_model.%s from *f90 files in the
    current directory.

    Additional Parameters ::
        -d/--debug
            Turn on assertion statements in F90 code

        -n/--no-compiler-optimization
            Do not send optimizing flags to compiler.
                 """ % ('pyd' if os.name == 'nt' else 'so')

usage['help'] = """kmcos help <command>
    Print usage information for the given command.
                """

usage['export'] = """kmcos export <xml-file> [<export-path>]
    Take a kmcos xml-file and export all generated
    source code to the export-path. There try to
    build the kmc_model.%s.

    Additional Parameters ::

        -s/--source-only
            Export source only and don't build binary

        -b/--backend (local_smart|lat_int)
            Choose backend. Default is "local_smart".
            lat_int is EXPERIMENTAL and not made
            for production, yet.

        -d/--debug
            Turn on assertion statements in F90 code.
            (Only active in compile step)

           --acf
            Build the modules base_acf.f90 and proclist_acf.f90. Default is false.
            This both modules contain functions to calculate ACF (autocorrelation function) and MSD (mean squared displacement). 
            
        -n/--no-compiler-optimization
            Do not send optimizing flags to compiler.
                    """ % ('pyd' if os.name == 'nt' else 'so')
          
usage['settings-export'] = """kmcos settings-export <xml-file> [<export-path>]
    Take a kmcos xml-file and export kmc_settings.py
    to the export-path.
                    """

usage['edit'] = """kmcos edit <xml-file>
    Open the kmcos xml-file in a GUI to edit
    the model.
                """

usage['import'] = """kmcos import <xml-file>
    Take a kmcos xml-file and open an ipython shell
    with the project_tree imported as pt.
                  """
usage['rebuild'] = """kmcos rebuild
    Export code and rebuild binary module from XML
    information included in kmc_settings.py in
    current directory.

    Additional Parameters ::
        -d/--debug
            Turn on assertion statements in F90 code
                    """

usage['shell'] = """kmcos shell
    Open an interactive shell and create a KMC_Model in it
               run == shell
               """
usage['run'] = """kmcos run
    Open an interactive shell and create a KMC_Model in it
               run == shell
               """

usage['version'] = """kmcos version
    Print version number and exit.
                   """

usage['view'] = """kmcos view
    Take a kmc_model.%s and kmc_settings.py in the
    same directory and start to simulate the
    model visually.

    Additional Parameters ::
        -v/--steps-per-frame <number>
            Number of steps per frame

                 """ % ('pyd' if os.name == 'nt' else 'so')

usage['xml'] = """kmcos xml
    Print xml representation of model to stdout
               """


def get_options(args=None, get_parser=False):
    import optparse
    import os
    from glob import glob
    import kmcos

    parser = optparse.OptionParser(
        'Usage: %prog [help] ('
        + '|'.join(sorted(usage.keys()))
        + ') [options]',
        version=kmcos.__version__)

    parser.add_option('-s', '--source-only',
                      dest='source_only',
                      action='store_true',
                      default=False)

    parser.add_option('-p', '--path-to-f2py',
                      dest='path_to_f2py',
                      default='f2py')

    parser.add_option('-b', '--backend',
                      dest='backend',
                      default='local_smart')
    parser.add_option('-a', '--avoid-default-state',
                      dest='avoid_default_state',
                      action='store_true',
                      default=False,
                      )

    parser.add_option('-v', '--steps-per-frame',
                      dest='steps_per_frame',
                      type='int',
                      default='50000')

    parser.add_option('-d', '--debug',
                      default=False,
                      dest='debug',
                      action='store_true')

    parser.add_option('-n', '--no-compiler-optimization',
                      default=False,
                      dest='no_optimize',
                      action='store_true')

    parser.add_option('-o', '--overwrite',
                      default=False,
                      action='store_true')

    parser.add_option('-l', '--variable-length',
                      dest='variable_length',
                      default=95,
                      type='int')

    parser.add_option('-c', '--catmap',
                      default=False,
                      action='store_true')
    
    parser.add_option('--acf',
                      dest='acf',
                      action='store_true',
                      default=False,
                      )
   
    try:
        from numpy.distutils.fcompiler import get_default_fcompiler
        from numpy.distutils import log
        log.set_verbosity(-1, True)
        fcompiler = get_default_fcompiler()
    except:
        fcompiler = 'gfortran'

    parser.add_option('-f', '--fcompiler',
                      dest='fcompiler',
                      default=os.environ.get('F2PY_FCOMPILER', fcompiler))

    if args is not None:
        options, args = parser.parse_args(args.split())
    else:
        options, args = parser.parse_args()
    if len(args) < 1:
        parser.error('Command expected')
    if get_parser:
        return options, args, parser
    else:
        return options, args


def match_keys(arg, usage, parser):
    """Try to match part of a command against
       the set of commands from usage. Throws
       an error if not successful.

    """
    possible_args = [key for key in usage if key.startswith(arg)]
    if len(possible_args) == 0:
        parser.error('Command "%s" not understood.' % arg)
    elif len(possible_args) > 1:
        parser.error(('Command "%s" ambiguous.\n'
                      'Could be one of %s\n\n') % (arg, possible_args))
    else:
        return possible_args[0]


def main(args=None):
    """The CLI main entry point function.

    The optional argument args, can be used to
    directly supply command line argument like

    $ kmcos <args>

    otherwise args will be taken from STDIN.

    """

    from glob import glob

    options, args, parser = get_options(args, get_parser=True)

    global model, pt, np, cm_model

    if not args[0] in list(usage.keys()):
        args[0] = match_keys(args[0], usage, parser)

    if args[0] == 'benchmark':
        from sys import path
        path.append(os.path.abspath(os.curdir))
        nsteps = 1000000
        from time import time
        from kmcos.run import KMC_Model
        model = KMC_Model(print_rates=False, banner=False)
        time0 = time()
        try:
            model.proclist.do_kmc_steps(nsteps)
        except:  # kmcos < 0.3 had no model.proclist.do_kmc_steps
            model.do_steps(nsteps)

        needed_time = time() - time0
        print('Using the [%s] backend.' % model.get_backend())
        print('%s steps took %.2f seconds' % (nsteps, needed_time))
        print('Or %.2e steps/s' % (1e6 / needed_time))
        model.deallocate()
    elif args[0] == 'build':
        from kmcos.utils import build
        build(options)
    elif args[0] == 'edit':
        from kmcos import gui
        gui.main()
    elif args[0] == 'settings-export':
        import kmcos.types
        import kmcos.io
        from kmcos.io import ProcListWriter

        if len(args) < 2:
            parser.error('XML file and export path expected.')
        if len(args) < 3:
            out_dir = '%s_%s' % (os.path.splitext(args[1])[0], options.backend)
            print('No export path provided. Exporting to %s' % out_dir)
            args.append(out_dir)

        xml_file = args[1]
        export_dir = args[2]
        project = kmcos.types.Project()
        project.import_file(xml_file)

        writer = ProcListWriter(project, export_dir)
        writer.write_settings()

    elif args[0] == 'export':
        import kmcos.types
        import kmcos.io
        from kmcos.utils import build
        if len(args) < 2:
            parser.error('XML file and export path expected.')
        if len(args) < 3:
            out_dir = '%s_%s' % (os.path.splitext(args[1])[0], options.backend)

            print('No export path provided. Exporting to %s' % out_dir)
            args.append(out_dir)

        xml_file = args[1]
        export_dir = os.path.join(args[2], 'src')

        project = kmcos.types.Project()
        project.import_file(xml_file)

        project.shorten_names(max_length=options.variable_length)

        kmcos.io.export_source(project,
                              export_dir,
                              options=options)

        if ((os.name == 'posix'
           and os.uname()[0] in ['Linux', 'Darwin'])
           or os.name == 'nt') \
           and not options.source_only:
            os.chdir(export_dir)
            build(options)
            for out in glob('kmc_*'):
                if os.path.exists('../%s' % out) :
                    if options.overwrite :
                        overwrite = 'y'
                    else:
                        overwrite = input(('Should I overwrite existing %s ?'
                                               '[y/N]  ') % out).lower()
                    if overwrite.startswith('y') :
                        print('Overwriting {out}'.format(**locals()))
                        os.remove('../%s' % out)
                        shutil.move(out, '..')
                    else :
                        print('Skipping {out}'.format(**locals()))
                else:
                    shutil.move(out, '..')

    elif args[0] == 'settings-export':
        import kmcos.io
        pt = kmcos.io.import_file(args[1])
        if len(args) < 3:
            out_dir = os.path.splitext(args[1])[0]
            print('No export path provided. Exporting kmc_settings.py to %s'
                  % out_dir)
            args.append(out_dir)

        if not os.path.exists(args[2]):
            os.mkdir(args[2])
        elif not os.path.isdir(args[2]):
            raise UserWarning("Cannot overwrite %s; Exiting;" % args[2])
        writer = kmcos.io.ProcListWriter(pt, args[2])
        writer.write_settings()

    elif args[0] == 'help':
        if len(args) < 2:
            parser.error('Which help do you  want?')
        if args[1] == 'all':
            for command in sorted(usage):
                print(usage[command])
        elif args[1] in usage:
            print('Usage: %s\n' % usage[args[1]])
        else:
            arg = match_keys(args[1], usage, parser)
            print('Usage: %s\n' % usage[arg])

    elif args[0] == 'import':
        import kmcos.io
        if not len(args) >= 2:
            raise UserWarning('XML file name expected.')
        pt = kmcos.io.import_xml_file(args[1])
        if len(args) == 2:
            sh(banner='Note: pt = kmcos.io.import_xml(\'%s\')' % args[1])
        elif len(args) == 3: # if optional 3rd argument is given, store model there and exit
            pt.save(args[2])

    elif args[0] == 'rebuild':
        from time import sleep
        print('Will rebuild model from kmc_settings.py in current directory')
        print('Please do not interrupt,'
              ' build process, as you will most likely')
        print('loose the current model files.')
        sleep(2.)
        from sys import path
        path.append(os.path.abspath(os.curdir))
        from tempfile import mktemp
        if not os.path.exists('kmc_model.so') \
           and not os.path.exists('kmc_model.pyd'):
            raise Exception('No kmc_model.so found.')
        if not os.path.exists('kmc_settings.py'):
            raise Exception('No kmc_settings.py found.')

        from kmcos.run import KMC_Model

        model = KMC_Model(print_rates=False, banner=False)
        tempfile = mktemp()
        f = file(tempfile, 'w')
        f.write(model.xml())
        f.close()

        for kmc_model in glob('kmc_model.*'):
            os.remove(kmc_model)
        os.remove('kmc_settings.py')
        main('export %s -b %s .' % (tempfile, options.backend))
        os.remove(tempfile)
        model.deallocate()

    elif args[0] in ['run', 'shell']:
        from sys import path
        path.append(os.path.abspath(os.curdir))
        from kmcos.run import KMC_Model

        # useful to have in interactive mode
        import numpy as np
        try:
            from matplotlib import pyplot as plt
        except:
            plt = None

        if options.catmap:
            import catmap
            import catmap.cli.kmc_runner
            seed = catmap.cli.kmc_runner.get_seed_from_path('.')
            cm_model = catmap.ReactionModel(setup_file='{seed}.mkm'.format(**locals()))
            catmap_message = '\nSide-loaded catmap_model {seed}.mkm into cm_model = ReactionModel(setup_file="{seed}.mkm")'.format(**locals())
        else:
            catmap_message = ''

        try:
            model = KMC_Model(print_rates=False)
        except:
            print("Warning: could not import kmc_model!"
                  " Please make sure you are in the right directory")
        sh(banner='Note: model = KMC_Model(print_rates=False){catmap_message}'.format(**locals()))
        try:
            model.deallocate()
        except:
            print("Warning: could not deallocate model. Was is allocated?")

    elif args[0] == 'version':
        from kmcos import VERSION
        print(VERSION)

    elif args[0] == 'view':
        from sys import path
        path.append(os.path.abspath(os.curdir))
        from kmcos import view
        view.main(steps_per_frame=options.steps_per_frame)

    elif args[0] == 'xml':
        from sys import path
        path.append(os.path.abspath(os.curdir))
        from kmcos.run import KMC_Model
        model = KMC_Model(banner=False, print_rates=False)
        print(model.xml())

    else:
        parser.error('Command "%s" not understood.' % args[0])


def sh(banner):
    """Wrapper around interactive ipython shell
    that factors out ipython version depencies.

    """

    from distutils.version import LooseVersion
    import IPython
    if hasattr(IPython, 'release'):
        try:
            from IPython.terminal.embed import InteractiveShellEmbed
            InteractiveShellEmbed(banner1=banner)()

        except ImportError:
            try:
                from IPython.frontend.terminal.embed \
                    import InteractiveShellEmbed
                InteractiveShellEmbed(banner1=banner)()

            except ImportError:
                from IPython.Shell import IPShellEmbed
                IPShellEmbed(banner=banner)()
    else:
        from IPython.Shell import IPShellEmbed
        IPShellEmbed(banner=banner)()



import export_import_library as eil
try:
    import kmcos.snapshots_globals as sg
    import kmcos.snapshots as snapshots
    import kmcos.throttling_globals as tg
    import kmcos.throttling as throttling
except:
    import snapshots_globals as sg
    import snapshots
    import throttling_globals as tg
    import throttling

# This function will reset any values that will be clobbered by loading from
# disk. It should manipulate the global variables directly. If nothing should
# be done, the body of this function should be replaced with 'pass'.
def set_runfile_params(load_saved_parameters_on):

    # The ending simulation time
    tg.cutoff_time = 100.0 # New cutoff time for simulation

    # Temperature
    sg.model.parameters.T = 600

    # PRNG state
    random_seed = -731543675
    set_PRNG_state(load_saved_parameters_on, random_seed)

# This is a function to load all of the saved module data.
def load_saved_data():

    # File names for loading/saving parameters
    tg_load_file = 'test_reaction_throttling_parameters.txt'
    tg_save_file = 'test_reaction_throttling_parameters.txt'
    sg_load_file = 'test_reaction_snapshots_parameters.txt'
    sg_save_file = 'test_reaction_snapshots_parameters.txt'

    # Export/import module objects for saving/loading data
    tg_eil_object = eil.module_export_import(tg_save_file, tg_load_file, tg)
    sg_eil_object = eil.module_export_import(sg_save_file, sg_load_file, sg)

    # Load modules
    sg_eil_object.load_params()
    tg_eil_object.load_params()

    # Reset number of steps and time
    sg.model.base.set_kmc_time(sg.kmc_time)
    sg.model.base.set_kmc_step(sg.steps_so_far)
    sg.atoms.kmc_step = sg.steps_so_far

    # Load the lattice
    sg.model._set_configuration(np.array(sg.config))
    sg.model._adjust_database()

    # Update the snapshot number
    tg.current_snapshot += 1

# This function initializes all the global variables for the runfile other than
# those required for the basic usage of do_throttled_snapshots.
def init(load_saved_parameters_on=False):

    # Load any saved data
    if load_saved_parameters_on:
        load_saved_data()

    # Reset any global variables that got clobbered
    set_runfile_params(load_saved_data_on)

def set_PRNG_state(load_saved_parameters_on, random_seed=None):

    if load_saved_parameters_on:
        snapshots.seed_PRNG(restart=True, state=sg.PRNG_state)
    else:
        snapshots.seed_PRNG(restart=False, state=random_seed)



#!/usr/bin/env python3
"""Very simple module that keeps several species
commonly needed to model heterogeneous catalyst kinetics
"""
from __future__ import print_function
#    Copyright 2009-2013 Max J. Hoffmann (mjhoffmann@gmail.com)
#    This file is part of kmcos.

#    kmcos is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    kmcos is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with kmcos.  If not, see <http://www.gnu.org/licenses/>.

import ase
import numpy as np
from numpy import interp as interp1d
from math import log
import os

janaf_data = None
try:
    import janaf_data
except:
    raise Exception("""
    Error: Could not import JANAF data
    Installing JANAF Thermochemical Tables
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    You can conveniently use gas phase chemical potentials
    inserted in rate constant expressions using
    JANAF Thermochemical Tables. A couple of molecules
    are automatically supported. If you need support
    for more gas-phase species, drop me a line.

    The tabulated values are not distributed since
    the terms of distribution do not permit this.
    Fortunately manual installation is easy.
    Just create a directory called `janaf_data`
    anywhere on your python path. To see the directories on your python
    path run ::

        python -c"import sys; print(sys.path)"

        Inside the `janaf_data` directory has to be a file
        named `__init__.py`, so that python recognizes it as a module ::

            touch __init__.py

            Then copy all needed data files from the
            `NIST website <http://kinetics.nist.gov/janaf/>`_
            in the tab-delimited text format
            to the `janaf_data` directory.""")


class Species(object):
    def __init__(self, atoms, gas=False, janaf_file='', name=''):
        self.atoms = atoms
        self.gas = gas
        if name:
            self.name = name
        else:
            if hasattr(self.atoms, 'get_chemical_formula'):
                self.name = self.atoms.get_chemical_formula(mode='hill')
            else:
                self.atoms.get_name()
        self.janaf_file = janaf_file

        # prepare chemical potential
        if self.gas and self.janaf_file and janaf_data is not None:
            self._prepare_G_p0(
                os.path.abspath(os.path.join(
                janaf_data.__path__[0],
                self.janaf_file)))

    def __repr__(self):
        return self.name

    def mu(self, T, p):
        """Expecting T in Kelvin, p in bar"""
        if self.gas:
            kboltzmann_in_eVK = 8.6173324e-5
            # interpolate given grid
            try:
                val = interp1d(T, self.T_grid, self.G_grid) + \
                       kboltzmann_in_eVK * T * log(p)
            except Exception:
                raise Exception('Could not find JANAF tables for %s.'
                                % self.name)
            else:
                return val

        else:
            raise UserWarning('%s is no gas-phase species.' % self.name)

    def _prepare_G_p0(self, filename):
        # from CODATA 2010
        Jmol_in_eV = 1.03642E-5
        # load data
        try:
            data = np.loadtxt(filename, skiprows=2, usecols=(0, 2, 4))
        except IOError:
            print('Warning: JANAF table %s not installed' % filename)
            return

        # define data
        self.T_grid = data[:, 0]
        self.G_grid = (1000 * (data[:, 2] - data[0, 2])
                               - data[:, 0] * data[:, 1]) * Jmol_in_eV

    def __eq__(self, other):
        return self.atoms == other.atoms and self.gas == other.gas

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return id(self)


# prepare all required species
H2gas = Species(ase.atoms.Atoms('H2', [[0, 0, 0], [0, 0, 1.2]],),
    gas=True,
    janaf_file='H-050.txt',
    name='H2gas')

H = Species(ase.atoms.Atoms('H'))

CH4gas = Species(ase.atoms.Atoms('CH4',
     [[-2.14262, 3.03116, 0.00000],
     [-1.07262, 3.03116, 0.00000],
     [-2.49979, 4.03979, 0.00000],
     [-2.51306, 2.50700, 0.85611],
     [-2.49435, 2.53348, -0.87948]],
     ), gas=True,
     janaf_file='C-067.txt',
     name='CH4gas')
CH4 = Species(ase.atoms.Atoms('CH4',
     [[-2.14262, 3.03116, 0.00000],
     [-1.07262, 3.03116, 0.00000],
     [-2.49979, 4.03979, 0.00000],
     [-2.51306, 2.50700, 0.85611],
     [-2.49435, 2.53348, -0.87948]],
     ),
     name='CH4')
O = Species(ase.atoms.Atoms('O',
    [[0, 0, 0]],
    cell=[10, 10, 10],
    ),
    name='O')
O2gas = Species(ase.atoms.Atoms('O2',
    [[0, 0, 0],
    [0, 0, 1.2]],
    cell=[10, 10, 10],
    ),
    gas=True,
    janaf_file='O-029.txt',
    name='O2gas')

NOgas = Species(ase.atoms.Atoms('NO',
    [[0, 0, 0], [0, 0, 1.2]],
    cell=[10, 10, 10],
    ),
    gas=True,
    janaf_file='N-005.txt',
    name='NOgas',
    )

NO = Species(ase.atoms.Atoms('NO', [[0, 0, 0], [0, 0, 1.2]],
                             cell=[10, 10, 10], ),
                name='NO',
                janaf_file='N-005.txt',
    )

NO2gas = Species(ase.atoms.Atoms(),
                 gas=True,
                 janaf_file='N-007.txt',
                 name='NO2gas')

NO3gas = Species(ase.atoms.Atoms(),
                 gas=True,
                 janaf_file='N-009.txt',
                 name='NO3gas')

COgas = Species(ase.atoms.Atoms('CO', [[0, 0, 0], [0, 0, 1.2]],
                                cell=[10, 10, 10],),
    gas=True,
    janaf_file='C-093.txt',
    name='COgas')
CO = Species(ase.atoms.Atoms('CO', [[0, 0, 0], [0, 0, 1.2]],
                             cell=[10, 10, 10], ),
    name='CO')

CO2gas = Species(ase.atoms.Atoms('CO2',
    [[0, 0, -1.2],
     [0, 0, 0],
     [0, 0, 1.2]],
     cell=[10, 10, 10],
     ),
     gas=True,
     janaf_file='C-095.txt',
     name='CO2gas')

NH3gas = Species(ase.atoms.Atoms(symbols='NH3',
          pbc=np.array([True,  True,  True], dtype=bool),
          cell=np.array(
      [[10.,   0.,   0.],
       [0.,  10.,   0.],
       [0.,   0.,  10.]]),
          positions=np.array(
      [[0.13288865,  0.13288865,  0.13288865],
       [-0.03325795, -0.03325795,  1.13361278],
       [-0.03325795,  1.13361278, -0.03325795],
       [1.13361278, -0.03325795, -0.03325795]])),
      gas=True,
      janaf_file='H-083.txt',
      name='NH3gas')

C2H4gas = Species(ase.atoms.Atoms(),
                  gas=True,
                  janaf_file='C-128.txt',
                  name='C2H4gas')

HClgas = Species(ase.atoms.Atoms(),
                 gas=True,
                 janaf_file='Cl-026.txt',
                 name='HClgas')

Cl2gas = Species(ase.atoms.Atoms(),
                 gas=True,
                 janaf_file='Cl-073.txt',
                 name='Cl2gas',)

H2Ogas = Species(ase.atoms.Atoms(),
                 gas=True,
                 janaf_file='H-064.txt',
                 name='H2Ogas',)

H2Oliquid = Species(ase.atoms.Atoms(),
                    gas=False,
                    janaf_file='H-063.txt',
                    name='H2Oliquid',)

